import sys

import typer

from .nautikos import Nautikos

app = typer.Typer()

nautikos = Nautikos()


@app.command()
def main(
    repository: str,
    tag: str,
    env: str = typer.Option(None),
    labels: str = typer.Option(None),
    config: str = typer.Option("nautikos.yaml"),
    dry_run: bool = typer.Option(False, "--dry-run"),
):
    s = f"Updating '{repository}' to '{tag}'"
    if env:
        s += f" in '{env}'"
    else:
        s += " in all environments"
    if labels:
        s += f" with labels '{labels}'"
    print(s)

    nautikos.set_dry_run(dry_run)
    nautikos.load_config(config)
    nautikos.update_manifests(
        repository, tag, environment=env, labels=labels.split(",") if labels else None
    )

    # Print modified files
    count_updated_img = 0
    for mod in nautikos.modifications:
        print(mod)
        if mod.updated:
            count_updated_img += 1

    # Determine output
    if len(nautikos.modifications) == 0:
        exit_msg = "ERROR - Didn't find any images to modify"
        exit_code = 1
    else:
        exit_msg = f"Updated {count_updated_img} out of {len(nautikos.modifications)} discovered occurances of '{repository}' to '{tag}'"  # noqa: E501
        exit_code = 0
    print(exit_msg)
    sys.exit(exit_code)



from __future__ import annotations

import pathlib
from typing import TypedDict

from .manifests import Modification, get_manifest
from .yaml import yaml


class ManifestConfig(TypedDict):
    path: str
    type: str
    labels: list[str]
    repositories: list[str]


class EnvironmentConfig(TypedDict):
    name: str
    manifests: list[ManifestConfig]


class ConfigData(TypedDict):
    environments: list[EnvironmentConfig]


class Nautikos:
    def __init__(self) -> None:
        self._workdir: pathlib.Path = pathlib.Path(".")
        self._dry_run: bool = False
        self._environments: list[EnvironmentConfig] = []
        self._modifications: list[Modification] = []

    @property
    def modifications(self) -> list[Modification]:
        return self._modifications

    def set_dry_run(self, dry_run: bool) -> None:
        self._dry_run = dry_run

    def load_config(self, path: str) -> None:
        self._workdir = pathlib.Path(path).parent
        with open(path, "r") as f:
            config_data: ConfigData = yaml.load(f)
        self._environments = config_data["environments"]

    def update_manifests(
        self,
        repository: str,
        new_tag: str,
        environment: str | None = None,
        labels: list[str] | None = None,
    ) -> None:
        """Updates image tags of given repository to new tag

        If environment is passed, only environments with matching name are modified;
        otherwise all environments are modified.

        If label is passed, only manifests with matching label are modified; otherwise
        all manifests in selected environments are modified.
        """
        # Get all relevant environments
        environments = self._get_environments(environment)

        # Get all relevant manifests
        manifests: list[ManifestConfig] = []
        for env in environments:
            manifests += self._get_manifests(env, labels)

        # Modify manifests
        for manifest_config in manifests:
            manifest = get_manifest(
                manifest_config["path"],
                manifest_config["type"],
                workdir=self._workdir,
            )
            manifest.load()
            manifest.modify(repository, new_tag)
            if len(manifest.modifications) > 0:
                if not self._dry_run:
                    manifest.write()
            self._modifications += manifest.modifications

    def _get_environments(self, environment: str | None) -> list[EnvironmentConfig]:
        envs: list[EnvironmentConfig] = []
        for env in self._environments:
            if not environment or env["name"] == environment:
                envs.append(env)
        return envs

    def _get_manifests(
        self, env: EnvironmentConfig, labels: list[str] | None = None
    ) -> list[ManifestConfig]:
        manifests: list[ManifestConfig] = []
        for manifest in env["manifests"]:
            if not labels or (
                "labels" in manifest and set(labels).issubset(set(manifest["labels"]))
            ):
                manifests.append(manifest)
        return manifests



import abc
import pathlib
from dataclasses import dataclass
from typing import Any, TypedDict

from .yaml import yaml


class Image(TypedDict):
    repository: str
    tag: str | None


@dataclass
class Modification:
    path: str
    repository: str
    previous: str
    new: str

    @property
    def updated(self) -> bool:
        return self.previous != self.new

    def __str__(self) -> str:
        if self.updated:
            return f"{self.path} -> modified '{self.repository}' to '{self.new}' (was '{self.previous}')"  # noqa: E501
        else:
            return f"{self.path} -> '{self.repository}' already up-to-date"


class AbstractManifest(abc.ABC):
    def __init__(self, path: str | pathlib.Path) -> None:
        self._path = path
        self._modifications: list[Modification] = []

    @property
    def modifications(self) -> list[Modification]:
        return self._modifications

    def load(self) -> None:
        with open(self._path, "r") as f:
            self._data = yaml.load(f)

    def write(self) -> None:
        with open(self._path, "w") as f:
            yaml.dump(self._data, f)

    @property
    def data(self) -> Any:
        if self._data:
            return self._data
        else:
            raise Exception("You must first load a manifest")

    def _record_modification(
        self, repository: str, old_tag: str | None, new_tag: str
    ) -> None:
        if not old_tag:
            old_tag = ""
        m = Modification(
            path=str(self._path),
            repository=repository,
            previous=str(old_tag),
            new=str(new_tag),
        )
        self._modifications.append(m)

    @abc.abstractmethod
    def modify(self, repository: str, new_tag: str) -> None:
        ...


KubernetesImageDefinition = str


class KubernetesContainer(TypedDict):
    image: KubernetesImageDefinition


class KubernetesManifest(AbstractManifest):
    def modify(self, repository: str, new_tag: str) -> None:
        for container in self._get_containers():
            parsed_image = self._parse_image(container["image"])
            if repository == parsed_image["repository"]:
                container["image"] = self._unparse_image(
                    {"repository": repository, "tag": new_tag}
                )
                self._record_modification(repository, parsed_image["tag"], new_tag)

    def _get_containers(self) -> list[KubernetesContainer]:
        return self.data["spec"]["template"]["spec"]["containers"]

    def _parse_image(self, image: KubernetesImageDefinition) -> Image:
        if ":" in image:
            repository, tag = tuple(image.split(":"))
        else:
            repository, tag = image, None
        return {"repository": repository, "tag": tag}

    def _unparse_image(self, image: Image) -> str:
        return f"{image['repository']}:{image['tag']}"


class KustomizeImageDefinition(TypedDict):
    name: str
    newTag: str


class KustomizeManifest(AbstractManifest):
    def modify(self, repository: str, new_tag: str) -> None:
        if "images" in self._data:
            for kustomize_image in self.data["images"]:
                if repository == kustomize_image["name"]:
                    old_tag = kustomize_image["newTag"]
                    kustomize_image["newTag"] = new_tag
                    self._record_modification(repository, old_tag, new_tag)

    def _parse_image(self, image: KustomizeImageDefinition) -> Image:
        return {"repository": image["name"], "tag": image["newTag"]}


def get_manifest(
    path: str | pathlib.Path, type: str, workdir: str | pathlib.Path | None = None
) -> AbstractManifest:
    if workdir:
        path = pathlib.Path(workdir) / pathlib.Path(path)
    if type == "kubernetes":
        return KubernetesManifest(path)
    elif type == "kustomize":
        return KustomizeManifest(path)
    elif type == "helm":
        raise Exception("Helm manifests are not yet implemented.")
    else:
        raise Exception(f"'{type}' is not a correct manifest type.")
